/*= main.c =====================================================================
 *
 * Copyright (c) 2006, Brge Strand
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  - Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 *  - Neither the name Brge Strand, his emloyer nor the names of its 
 *    contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * 
 * Main executable program for DSA communication. Tested with Philips L1210/65
 * CD drive
 * 
 * Version 0.1
 *
 * Tested with avr-gcc (GCC) 3.4.6
 *==============================================================================
 */


// DSA todo:
// - Clean up code with pointers instead of global variables
// - Make startup of CD drive take a bit less time
// - Make code compileable without #define DEBUG
// - Check out undocumented features (like lacking 0xAA at end of last track and 0x8101 at reset)
// - Detect negative time before start of track
// - Fast forward and fast backward, supposed to be tricky on this player
// - Complete UI on MCU buttons. 
// - Updated display routines. First write decimal to terminal, then hook up a display
// - Implement RC 5 decoder

#include "defines.h"

char error;											// Global error status of some register accesses
char statusbyte;									// Global nRF24L01 status byte returned every SPI access
char wakeup;										// Global register saying whether MCU woke from internal timer or external interrupt
char wakeup_one_shot;								// Global register saying whether MCU woke up from one-shot timer or not
char rb0; 											// FIX: Not pretty, maybe go all out and user pointers. 
char rb1;
char mode = CD_PENDING;							// The operation mode, or where in the flowchar we are right now
	
int main(void) {

	char current_track = 0;                         // Local variables to keep track of what is playing and how CD is put together
	char current_min = 0;
	char current_sec = 0;
	char first_track = 0;
    char last_track = 0;
	char time_min = 0;
	char time_sec = 0;	
	
	int c0 = 0;										// Used to decode button info from uart

	mcu_init();
	mcu_watchdog(WD_ON);							// Use watchdog to keep track of playing CD

	#ifdef DEBUG									// Selectively run with debugger and interactive debugger
		mcu_uart_init();
		db_putenter();
		db_putenter();
		mcu_putchar('-');
		mcu_putchar('D');
		mcu_putchar('S');
		mcu_putchar('A');
		mcu_putchar('-');
		db_putenter();

		#ifdef DEBUG_IF
			db_hex();								// Run a debugging interface towards RS232 terminal
		#endif
	#endif

	mcu_reset_cd();									// Reset the CD drive before starting to give it commands
	db_display_CD(EMPTYCHAR, EMPTYCHAR, EMPTYCHAR); // Clear display when status of CD is unknown


// Which buttons will be considered in the different modes?
//
// CD_PENDING	 	-
// CD_READING_TOC	-
// CD_PLAYING		Skip f+b, scan f+b, Pause, Stop
// CD_STOPPED		Skip f+b, Play, Open
// CD_PAUSED	    Play/pause
// CD_OPENING		Open, Close
// CD_OPENED		Close
// CD_CLOSING		Open, Close
// CD_NO_DISC		Open, Close


// Watchdog interrupt will be active all the time, striking an interrupt every 2 seconds if not reset. The interrupt
// will only take effect if mode==CD_PLAYING. Then the watchdog will issue a DSA stop command. This is used to stop playback
// at end of the CD. But in order to not have rouge stop commands, the watchdog has to be reset within 2 second
// intervals as long as mode==CD_PLAYING

	while (1) {
		while ((mcu_ishigh_D()) & !(mcu_readkey())) { // Wait for a negative edge on D or user IO on keyboard
		}											// After this CD player has data to send or user has command

		//////////////////////////
		//					    //
		// Decode user buttions //
		//					    //
		//////////////////////////

		if (mcu_readkey()) {						// User pressed a key, now collect command from user
			c0 = mcu_getchar();
			if (c0=='f') {							// 'f' for skip forward
				if ((mode == CD_STOPPED) || (mode == CD_PLAYING)) { // Only skip forwards when there is a CD to do it with
					if (current_track < last_track) {	// If there exists a next track
						mcu_watchdog(WD_RESET);		// Watchdog interrupt shouldn't interrupt us
						current_track++;
						if (mode == CD_PLAYING) 
							mcu_write_DSA(0x01, current_track); // If we're playing, go to play next track
						current_min = 0;
						current_sec = 0;
						db_display_CD(current_track, EMPTYCHAR, EMPTYCHAR);	 // Alternatively, display track with empty min and sec fields before play
					}
				}
			}
			else if (c0=='t') {					// 't' for stop
				mode = CD_PENDING;					// Allow very little action until DSA system reports stopped
				mcu_write_DSA(0x02, 0x00);  
			}
			else if (c0=='b') {					// 'b' for skip backward
				if ((mode == CD_STOPPED) || (mode == CD_PLAYING)) { // Only skip backwards when there is a CD to do it with
					mcu_watchdog(WD_RESET);			// Watchdog interrupt shouldn't interrupt us
					if (current_min + current_sec == 0) {	// If we're not into a track,
						if (current_track > first_track)	// Select previous existing track. If into a track, don't select another
							current_track--;
					}
					if (mode == CD_PLAYING)
						mcu_write_DSA(0x01, current_track); // If we're playing, go to the track selected just above
					current_min = 0;
					current_sec = 0;
					db_display_CD(current_track, EMPTYCHAR, EMPTYCHAR);	 // Alternatively, display track with empty min and sec fields before play
				}
			}
			else if (c0=='F')	{		   			// 'F' for fast forward FIX: implement
				if (mode == CD_PLAYING) {
				
				}
			}
			else if (c0=='B')	{		   			// 'B' for fast backwards FIX: implement
				if (mode == CD_PLAYING) {
				
				}
			}
			else if (c0=='p') {					// 'p' to play, only works from CD_STOPPED. FIX: Integrate nicely with pause!
				if (mode == CD_STOPPED) {
					mcu_write_DSA(0x01, current_track);		// Play selected song (default is 1st) on CD. DSA will set mode=CD_PLAYING
				}
			}
			else if (c0=='P')	{		   			// 'P' for pause
				if (mode == CD_PLAYING) {
				    mcu_write_DSA(0x04, 0x00);		// Pause
					mode = CD_PAUSED;
					db_display_CD(current_track, current_min, EMPTYCHAR); // FIX: Blink display to indicate pause
				}
				else if (mode == CD_PAUSED) {
				    mcu_write_DSA(0x05, 0x00);		// Pause release
					mode = CD_PLAYING;
					mcu_watchdog(WD_RESET);			// Watchdog interrupt shouldn't interrupt us
					db_display_CD(current_track, current_min, current_sec); // Back to ordinary playing display
				}
			}
			else if (c0=='o')	{		   			// 'o' for open
				if ((mode == CD_STOPPED) || 		// Only if in one of several modes should we allow the CD tray to open
						(mode == CD_OPENING) || 
						(mode == CD_CLOSING) || 
						(mode == CD_NO_DISC)) {
					mode = CD_PENDING;				// Allow precious little while waiting for DSA system's response
					db_display_CD(EMPTYCHAR, EMPTYCHAR, EMPTYCHAR); // Clear display when opening tray
					mcu_write_DSA(0x6A, 0x00);		// Clear TOC
					mcu_write_DSA(0x0A, 0x00);		// Open the tray, DSA system will set mode=CD_OPENING, then CD_OPENED
				}
			}
			else if (c0=='c')	{		   			// 'c' for close
				if ((mode == CD_OPENING) || 		// Only if in one of several modes should we allow the CD tray to close
						(mode == CD_CLOSING) || 
						(mode == CD_OPENED) ||
						(mode == CD_NO_DISC)) {
					mode = CD_PENDING;				// Allow precious little while waiting for DSA system's response
					mcu_write_DSA(0x0B, 0x00);		// Close the tray, DSA system will set mode=CD_CLOSING, then CD_READING_TOC
				}
			}
		}


		//////////////////////////
		//					    //
		// Decode DSA feedback  //
		//					    //
		//////////////////////////

		if (mcu_islow_D()) {						// If oponent wants to send data to us,
			if (mcu_read_DSA()) {		   			// Try receiving that data, and if it comes through OK,
				switch (rb0) {
					case 0x01:						// 0x01 from CD drive means Found. Do nothing for now
					break;
					case 0x02:						// 0x02 from CD drive means Stopped. Do nothing for now
						mode = CD_STOPPED;
						db_display_CD(last_track, time_min, time_sec); // Show the contents of CD
					break;
					case 0x0B:						// Got Tray moving, now decode the direction, page 27
						if (rb1 == 0x01)
							mode = CD_OPENING;
						else if (rb1 == 0x00) {	// Tray could be closing because we told it to or because user pushed it.
							if (mode != CD_CLOSING)		// If user is pushing it, 
								mcu_write_DSA(0x0B, 0x00);	// Command CD module to close tray itself
							mode = CD_CLOSING;		// Anyway we're moving in
						}
					break;
					case 0x0C:						// Got Tray opened
						mode = CD_OPENED;
					break;
					case 0x0D:						// Got Tray closed
						mode = CD_READING_TOC;
						mcu_write_DSA(0x03, 0x00);	// Read the TOC after the tray closes. That will eventually give mode=CD_STOPPED or CD_NO_DISC
					break;
					case 0x10:						// 0x10 from CD drive updates "Actual title" = track number
						current_track = rb1;
					break;
					case 0x11:			  			// 0x11 from CD drive updates "Actual index" = not too much use
					break;
					case 0x12:			  			// 0x12 from CD drive updates minutes
						current_min = rb1;
					break;
					case 0x13:			  			// 0x13 from CD drive updates seconds
						mcu_watchdog(WD_RESET);		// We're playing music, don't let the watchdog interrupt that before the seonds update ceases
						mode = CD_PLAYING;			// Assume that the seconds updating means we're playing. FIX: use found?
						current_sec = rb1;
						db_display_CD(current_track, current_min, current_sec); // FIX: add "-" indicator
					break;
					case 0x20:			  			// 0x20 from CD tells which track is the first one
						first_track = rb1;
					break;
					case 0x21:			  			// 0x21 from CD tells which track is the last one
						last_track = rb1;
					break;
					case 0x22:			  			// 0x22 from CD tells total time in minutes
						time_min = rb1;
					break;
					case 0x23:			  			// 0x23 from CD tells total time in seconds. This is usually the last part of TOC...
						time_sec = rb1;	 			// ...So for now print the TOC to the user
						current_track = first_track;	// Allow user interface to first skip and then play
						mcu_write_DSA(0x02, 0x00);	// Send a stop command to stop the disk after reading TOC. That should soon give mode=CD_STOPPED
					break;
					case 0x6A:						// Got TOC_cleared 0x6A
						current_track = 0;			// Forget all about the CD that used to be loaded
						current_min = 0;
						current_sec = 0;
						first_track = 0;
						last_track = 0;
						time_min = 0;
						time_sec = 0;				
					break;
					case 0x81:			  			// We got the undocumented 0x81 that the drive reports after reset. Use that to read TOC
						mode = CD_READING_TOC;
						mcu_write_DSA(0x03, 0x00);	// FIX: this is hardly the right way to do it....
					break;
					case 0x04:						// Detected an error. Now decode the different ones
						if (rb1 == 0x02)			// Got Focus error, or no disc
						    mode = CD_NO_DISC;
						else {
							mcu_putchar('r');		// Print the error
							db_puthex(rb0);	
							db_puthex(rb1);
							db_putenter();
						}
					break;
					default:						// We got something that we don't yet know how to decode...
						mcu_putchar('r');
						db_puthex(rb0);	 			// Print it out.
						db_puthex(rb1);
						db_putenter();
					break;
				} // end of switch 
			}
			else
				mcu_putchar('$');		   			// Reception failed after a low D was detected
		}
	} 

} // main


