/*= mcu_atmega48.c =============================================================
 *
 * Copyright (c) 2006, Brge Strand
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  - Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 *  - Neither the name Brge Strand, his emloyer nor the names of its 
 *    contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * 
 * Hardware dependant implementation of mcu code. This is where to put any
 * hardware dependant #include statements. This file also contains the IO list
 * that the code uses.
 * 
 * Version 0.1
 *
 * Tested with avr-gcc (GCC) 3.4.6
 *==============================================================================
 */

// MCU specific code for Atmel ATmega48
// This file must be ported to any other MCU to be used. 
// It implements all functions in mcu.h
// All defines specific to ATmega48 reside in this file!

// The following hardware is used in the ATmega48, 0=MCU input, 1=MCU output:

// PB0 - pin 14 - 1 CD reset output
// PB1 - pin 15 - 0 
// PB2 - pin 16 - 0 
// PB3 - pin 17 - 1 MOSI, ISP, RFIO
// PB4 - pin 18 - 0 MISO, ISP, RFIO
// PB5 - pin 19 - 1 SCK, ISP, RFIO
// PB6 - pin 9  - 0 XO pin, lay off on STK500
// PB7 - pin 10 - 0 XO pin, lay off on STK500

// PC0 - pin 23 - 1 A out CD     The external D/S/A signal is connected to D/S/A in. Three 47k pull up resistor go from
// PC1 - pin 24 - 1 S out CD     D/S/A in to VDD. Three 1N4148 diodes go from D/S/A in to D/S/A out.
// PC2 - pin 25 - 1 D out CD             .----|___|--- Vdd=5V
// PC3 - pin 26 - 0 A in CD              |      __________
// PC4 - pin 27 - 0 S in CD       CD ----.-----| in
// PC5 - pin 28 - 0 D in CD            __|__   |
// PC6 - pin 1  - 0 Reset, ISP         _\ /_   |      MCU
//                                       |     |
// PD0 - pin 2  - 0 RXD, UART debug      .-----| out
// PD1 - pin 3  - 1 TXD, UART debug            |__________
// PD2 - pin 4  - 0 INT0
// PD3 - pin 5  - 0
// PD4 - pin 6  - 0
// PD5 - pin 11 - 0
// PD6 - pin 12 - 1 debug output
// PD7 - pin 13 - 1 debug output

// Watchdog timer interrupt - Monitor end of last track
// Timer/Counter0			- 
// Timer/Counter1			- Wait for a number of ms
// Timer/Counter2			- Protect against DSA timeouts


#include "mcu.h"

// MCU specific includes, change if porting to a different MCU or compiler
#include <stdio.h>
#include <avr/interrupt.h> 
// #include <avr/signal.h>
#include <avr/sleep.h> 
#include <avr/wdt.h> 
#include <avr/pgmspace.h>
#include <avr/io.h>


// Some forward declarations of functions that are not available externally
void mcu_wait_timer(void);


// A local define for this MCU
#define F_CPU 1000000UL								// Internal RC oscillator, with 8x division in fuse bits (default for ATmega48)
//#define F_CPU 8000000UL							// Internal RC oscillator, without 8x division in fuse bits 


// Handler for timer/counter0 interrupt. Not used at the time
SIGNAL(SIG_OUTPUT_COMPARE0A) { 
} 


// Handler for timer/counter1 interrupt
SIGNAL(SIG_OVERFLOW1) { 
	// Simply return after interrupt
} 


// Handler for timer/counter2 interrupt. 
SIGNAL(SIG_OUTPUT_COMPARE2A) { 
	extern char wakeup_one_shot;					// Global flag that tracks of what woke up the one-shot timer

	TIMSK2 = 0b00000000;							// Disable all T/C 2 interrupts
	TCCR2B = 0b00000000;							// Stop clock by writing 0 to the clock source bits.
	wakeup_one_shot = WAKE_WAIT;					// Internal timer interrupt caused the wakeup
} 


// Handler for watchdog timer interrupt. This interrupt strikes for two reasons: 
// 1 - system is not in CD_PLAYING mode, in which it should have no affect
// 2 - system is in CD_PLAYING mode and has not got a seconds update for too long (last song). Then stop the disc. 
SIGNAL(SIG_WATCHDOG_TIMEOUT) {
	extern char mode;								// Global mode flag for the state machine

	// mcu_putchar('D');							// Indicate that the watchdog struck
	
	if (mode == CD_PLAYING)						// If interrupt strikes while playing music,
		mcu_write_DSA(0x02, 0x00);					// Stop playing music
}


// Handler for external interrupt
SIGNAL(SIG_INTERRUPT0) { 
} 


// Initialize the MCU. 
void mcu_init(void) {

	DDRB  = 0b00101001;
	PORTB = 0b00000000;								// Set CSN=1 and clear all other PORTB outputs while removing PINB pull-ups

	DDRC  = 0b00000111;
	PORTC = 0b00000111;								// Set D,S,A all high on startup

	DDRD  = 0b00000010;
	PORTD = 0b00000000;								// Clear all pins. No pull-ups on inputs

	SPCR = (1<<SPE)|(1<<MSTR);						// Enable SPI master at fck/4. 
	SPSR = 0x01;									// Enable SPI master at fck/2. Data sheet p. 166
	
	// FIX: Shut down analog comparator, brownout detector, spi etc. 

	cli();											// Setting up interrupts in atomic operation
	EIFR |= 0x01;									// Clear interrupt flag before looking for new one datasheet p. 66
	EICRA |= 0x02;									// Int on falling edge of INT0 datasheet p. 65
	EIMSK |= 0x01;									// Enable INT0 on bit 0 of EIMNSK datasheet p. 66
	sei();											// Enable interrupts thereafter

	mcu_wait_ms(32);								// Wait for things to stabilize
}	


// DSA pins FIX: convert to macros!

// DSA IO definitions, very hardware dependant
#define DINMASK 		0b00100000					// Pin location of D input
#define SINMASK 		0b00010000					// Pin location of S input
#define AINMASK 		0b00001000					// Pin location of A input
#define DOUTMASK 		0b00000100					// Pin location of D output
#define SOUTMASK 		0b00000010					// Pin location of S output
#define AOUTMASK 		0b00000001					// Pin location of A output

char mcu_ishigh_D(void) {
	return PINC & DINMASK;
}

char mcu_islow_D(void) {
	return !(PINC & DINMASK);
}

char mcu_ishigh_S(void) {
	return PINC & SINMASK;
}

char mcu_islow_S(void) {
	return !(PINC & SINMASK);
}

char mcu_ishigh_A(void) {
	return PINC & AINMASK;
}

char mcu_islow_A(void) {
	return !(PINC & AINMASK);
}

void mcu_set_D(void) {
	PORTC |= DOUTMASK;
}

void mcu_clear_D(void) {
	PORTC &= ~DOUTMASK;
}

void mcu_set_S(void) {
	PORTC |= SOUTMASK;
}

void mcu_clear_S(void) {
	PORTC &= ~SOUTMASK;
}

void mcu_set_A(void) {
	PORTC |= AOUTMASK;
}

void mcu_clear_A(void) {
	PORTC &= ~AOUTMASK;
}


// Send a reset pulse (active high) to the CD drive
void mcu_reset_cd(void) {
	PORTB |= 0x01;									// Set PB0 = reset to CD drive
	mcu_wait_ms(100);								// wait 100ms
	PORTB &= ~0x01;									// Clear PB0 = end of CD drive reset
}


// Set up timer/counter 0 for sample rate, not used yet
void mcu_startsamplecounter(void) {	
	TCCR0A = 0b00000010;							// Clear-on-compare mode
	TCCR0B = 0b00000100;							// Enable timer, prescale by 256
//	OCR0A  = 124;									// 8MHz / 256 / 125 is 250Hz, 4ms
	OCR0A  = 123;									// 8MHz / 256 / 124 is 252Hz, 3.968ms, shortest time for 25 packets. 4ms+130s is max cycle time
	TIMSK0 = 0b00000010;							// Interrupt at compare match
}


// Read a user input button
// FIX: convert between active high and active low buttons!
char mcu_getbutton(void) {
	if ((PINB & 0b00000001)==0x00)					// Look for Play at PINB0
		return BTN_PLAY;
	else if ((PINB & 0b00000010)==0x00)			// Look for Scan forwards at PINB1
		return BTN_SCANF;
	return BTN_NOKEY;
}


// Set up watchdog timer in interrupt mode for sample rate
void mcu_watchdog(char operation) {	
	if (operation == WD_ON) {
		cli();										// Disable all interrupts while manipulating watchdog
		wdt_reset();								// watchdog reset
		WDTCSR |= (1<<WDCE) | (1<<WDE);			// Start timed sequence, enable watchdog changes
		WDTCSR =  (1<<WDIE) | (1<<WDP3);			// Set up timeout value around 4 seconds. Hopefully, no seek operation lasts that long!
		sei();										// Enable interrupts again
	}
	else if (operation == WD_RESET) {					// Delay the strike of the watchdog!
		wdt_reset();
	}
	else if (operation == WD_OFF) {
		cli();										// Disable interrupts
		wdt_reset();								// Watchdog reset
		MCUSR &= ~(1<<WDRF);						// Set WDRF bit in MCUCR to 0 to clear watchdog interrupt flag, needed?
		WDTCSR |= (1<<WDCE) | (1<<WDE);			 // Enable changes in WDTCSR
		WDTCSR = 0x00;								// Turn off watchdog timer
		sei();										// Turn on interrupts again
	}	
}


// Halt the MCU for a number of 1ms periods by means of low-power and high-power sleeps
void mcu_wait_ms(unsigned int time) {
	extern char wakeup;

	wakeup = WAKE_WAIT;								// Assume that the delay is going to wake up the MCU

	while ((time > 0) && (wakeup == WAKE_WAIT)) {	// Wait until set time has passed, or until another interrupt arrives
		TCNT1 = -((uint16_t)(1e-3 * F_CPU));		// Sleep 1ms
		mcu_wait_timer();							// Do the actual sleep for 1ms
		time--;										// Prepare to wait a little more
	}
}


// Go in to sleep mode after the calling function has set the desired delay
// The only allowable interrupt source is timer interrupt!
void mcu_wait_timer(void) { 
	cli();											// Go for atomic operations while setting up the interrupt source
	TIMSK1 = (1<<TOIE1);							// Enable t/c 1 overflow int 
	TCCR1A = 0;										// t/c 1 normal mode 
	TCCR1B = (1<<CS10);							 // Start timer 1 with prescaler 1 
	set_sleep_mode(SLEEP_MODE_IDLE);				// Idle mode keeps counters running
	sei();											// Set global interrupt flag
	sleep_mode();									// Wait until interrupted by pin or counter
	cli();											// Clear global interrupt flag
	TIMSK1 = 0x00;									// Disable all t/c 1 interrupts
	TCCR1B = 0x00;									// Stop t/c 1 
	sei();											// Turn on interrupts again
}  


// Starts a wake period that will interrupt the MCU once a little later
void mcu_sleep_one_shot_enable(unsigned char time) {
	extern char wakeup_one_shot;					// Global flag that tracks of what woke up the one-shot timer

	// FIX: Make wait period independant of MCU speed
	// FIX: Make proper quantization in terms of ms

	wakeup_one_shot = WAKE_PENDING;					// We are waiting to wake up from a compare match on Timer / Counter 2, nothing else
	cli();											// Go for atomic operations while setting up the interrupt source
	TCCR2A = 0b00000000;							// Normal mode, T/C 2 counts from 0x00 to 0xFF. See p. 1|49
	TCCR2B = 0b00000100;							// Prescale by 64 for now. See p. 152
	TCNT2  = 0b00000000;							// Start counting from 0. See p. 153
	OCR2A  = time;									// The compare register. This one is set according to wait period.
	TIMSK2 = 0b00000010;							// Enable interrupt on compare match with OCR2A. No other T/C 2 interrupts
	sei();											// Turn on interrupts again
}


// Disable the one-shoot wait period
void mcu_sleep_one_shot_disable(void) {
	extern char wakeup_one_shot;
	
	TIMSK2 = 0b00000000;							// Disable all T/C 2 interrupts
	TCCR2B = 0b00000000;							// Stop clock by writing 0 to the clock source bits.
	wakeup_one_shot = WAKE_PENDING;					// We are not waiting anymore FIX: rather use WAKE_DISABLE or something?
}


// Turn a LED on or off
void mcu_led(char mode) {
}


// Debug on LED on PD6
void mcu_flip(void) {
	PORTD |= 0b01000000;
	PORTD &= 0b10111111;
}


// Debug on LED on PD7
void mcu_flop(void) {
	PORTD |= 0b10000000;
	PORTD &= 0b01111111;
}


#ifdef DEBUG

// Sends a single char over the uart
int mcu_putchar(char c) {
	loop_until_bit_is_set(UCSR0A,UDRE0);			// Wait until the send register is empty
	UDR0 = c;										// Send char to send register
	return 0;
}


// Is the user currently pressing a key?
char mcu_readkey(void) {
	if (UCSR0A & 1<<RXC0)							// Check the bit that indicates if data is being received on UART
		return 1;
	else
		return 0;
}


// Halts, then receives a single char from the uart. Immediate echo. FIX: implement some kind of sleep mode here!
int mcu_getchar(void) {
	char c;
	loop_until_bit_is_set(UCSR0A,RXC0);				// Wait untill a char is received
	c = UDR0;										// Read char from uart
	mcu_putchar (c);								// Echo the character   
	return c;
}

 
// Sets up the uart 8 data bits, 1 stop bit
void mcu_uart_init(void) {
	UBRR0L = 12;									// With 1MHz we have 4800 baud, with 8MHz we have 38400 baud
	UCSR0B = (1<<RXEN0)|(1<<TXEN0);				// Enable both read and write
	UCSR0C = (3<<UCSZ00);							// 8 bits data, 1 stop bit
}

#endif // DEBUG


// Wait for timeout (return 0) or an event (return 1). Event may be high or low D,S,A
char mcu_wait_for_event(char event) {  // FIX: also specify timeout period in parameter list
	extern char wakeup_one_shot;
	char temp = 3;

	mcu_sleep_one_shot_enable(200);						// Allow the CD transport plenty of time to respond
	while (temp==3) {
		if (wakeup_one_shot==WAKE_WAIT)				// Terminate loop by timeout...
			temp = 0;
		switch (event) {
			case ISHIGH_D:
				if (mcu_ishigh_D())					// .. or by a high D
					temp = 1;
			break;
			case ISHIGH_S:
				if (mcu_ishigh_S())					// .. or by a high S
					temp = 1;
			break;
			case ISHIGH_A:
				if (mcu_ishigh_A())					// .. or by a high A
					temp = 1;
			break;
			case ISLOW_D:
				if (mcu_islow_D())						// .. or by a low D
					temp = 1;
			break;
			case ISLOW_S:
				if (mcu_islow_S())						// .. or by a low S
					temp = 1;
			break;
			case ISLOW_A:
				if (mcu_islow_A())						// .. or by a low A
					temp = 1;
			break;
		}
	}
	mcu_sleep_one_shot_disable();
	
	return temp;
}

// Write a 16-bit word to DSA, MSB first. Write MSbit of b0 first, LSbit of b1 last
char mcu_write_DSA(char b0, char b1) {
	char bits_tx=0;
	char dtx = 0;  // FIX: merge with FIFO??

	mcu_clear_D();
	if (mcu_wait_for_event(ISLOW_A)) {
		mcu_set_D();
		if (mcu_wait_for_event(ISHIGH_A)) {

			// TX data:
			while (bits_tx < 16) {
				if (b0 & 0x80) {					// MSbit of MSbyte is output
					// mcu_putchar('1');
					// mcu_wait_ms(1);
					mcu_set_D();
				}
				else {
					// mcu_putchar('0');
					// mcu_wait_ms(1);
					mcu_clear_D();
				}
				
				b0 <<= 1;							// MSbyte is shifted left
				if (b1 & 0x80)						// MSbit of LSbyte is...
					b0 |= 0x01;						// ...shifted into LSbit of MSbyte
				b1 <<= 1;							// LSbyte is shifted left

				mcu_clear_S();
				if (mcu_wait_for_event(ISLOW_A)) {
					mcu_set_D();					// Won't get comm.ack from CD drive if D is stuck on low
					mcu_set_S();
					if (mcu_wait_for_event(ISHIGH_A)) {
						bits_tx++;
					}
					else {
						bits_tx = 18;
						mcu_putchar('i');
					}
				}
				else {
					bits_tx = 17;
					mcu_putchar('j');
				}
			} // end of while loop

			// TX comm.ack
			if (bits_tx == 16) {
				mcu_clear_A();
				if (mcu_wait_for_event(ISLOW_S)) {
					if (mcu_ishigh_D())
						dtx = 1;					// CD drive counted the bits correctly!
					mcu_set_A();
					if (mcu_wait_for_event(ISHIGH_S)) {
						// All is well, do nothing if this one pans out
					}
					else {
						dtx = 0;					// Missed last edge, things didn't work out after all
						mcu_putchar('m');
					}				
				}
				else {
					mcu_putchar('k');
				}			
			}
		}
		else {
			mcu_putchar('g');
			// Set DSA below
		}
	}
	else {
		mcu_putchar('h');
		// set DSA below
	}
	mcu_set_D();
	mcu_set_S();
	mcu_set_A();
	
	return dtx;										// Return 1 if CD drive sent comm.ack, 0 otherwise
}


// Read a 16-bit word from DSA, MSB first. Read MSbit of rb0 first, LSbit of rb1 last
char mcu_read_DSA(void) {
	extern char rb0; // FI0X: Not pretty, maybe go all out and user pointers. 
	extern char rb1;

	char bits_rx = 0;
	char drx = 0;	// FIX: make global!

	// We are here when D has gone low from the other side of DSA

	// RX sync FIX: Ignore test for drx==0
	mcu_clear_A();									// Acknowledge the transmit request from CD player
	if (mcu_wait_for_event(ISHIGH_D)) {			// If we get a high D before timeout..
		mcu_set_A();								// Confirm that main MCU wants to listen to CD player
				
		// RX data:
		while (bits_rx < 16) {
			if (mcu_wait_for_event(ISLOW_S)) {
					
				rb0 <<= 1;							// MSbyte is shifted left
				if (rb1 & 0x80)						// MSbit of LSbyte is...
					rb0 |= 0x01;					// ...shifted into LSbit of MSbyte
				rb1 <<= 1;							// LSbyte is shifted left
				if (mcu_ishigh_D())
					rb1 |= 0x01;					// LSbit of LSbyte is input

				mcu_clear_A();
				if (mcu_wait_for_event(ISHIGH_S)) {
					mcu_set_A();
					bits_rx++;						
				}
				else
					bits_rx = 18;					// Indicate a failure, waiting for Strobe high failed
			}
			else
				bits_rx = 17;						// Indicate a failure, waiting for Strobe low failed.
		}
	}
		
	mcu_set_D();									// Set DSA when wait for a high D fails and when it is time
	mcu_set_S();									// To start sending the comm.ack.
	mcu_set_A();

	// RX comm.ack:
	if (mcu_wait_for_event(ISLOW_A)) {
		if (bits_rx != 16) 
			mcu_clear_D();
		mcu_clear_S();
				
		if (mcu_wait_for_event(ISHIGH_A)) {
			drx = 1; // Signal to higher level code that we have to do something about some incoming data
		}
	}

	mcu_set_D();									// Set DSA when wait for a high D fails and when it is time
	mcu_set_S();									// To start sending the comm.ack.
	mcu_set_A();
	
	return drx;									// Returns 1 if data was picked up, 0 otherwise
}

