/*= uartdebug.c ================================================================
 *
 * Copyright (c) 2006, Brge Strand
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  - Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 *  - Neither the name Brge Strand, his emloyer nor the names of its 
 *    contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * 
 * Debug routines that lets an UART act as display and buttons for a CD player.
 * This file is hardware independant, so low-level UART routines are in 
 * mcu_xxx.c
 * 
 * Version 0.1
 *
 * Tested with avr-gcc (GCC) 3.4.6
 *==============================================================================
 */


#include "uartdebug.h"

#ifdef DEBUG										// All of this file is inside this ifdef!

// Some forward declarations of functions that are not available externally
int db_is_ascii_hex(int c);
char db_ascii_to_hex(int c0, int c1);

char db_getcommand(void);
void db_printstatus(void);
void db_rf_putbyte(void);
void db_rf_getbyte(void);


// Pick up an nRF24L01 command from the user interface
char db_getcommand(void) {
	extern char error;								// Global error register
	char temp;

	mcu_putchar('{');								// Print '{' before start of command
	// Get command as two hex numbers or terminate
	
	temp = db_gethex(mcu_getchar());				// Get an 8-bit hex number from terminal
	if (error == 0x00) {							// If valid hex number was received, finish address printout
		mcu_putchar('}');							// Print '}'
	}
	return temp;									// Always return the command, caller must test extern char error
}


// Print out the status byte
void db_printstatus(void) {
	extern char statusbyte;						// Global register for nRF24L01 satus byte
	
	mcu_putchar(' ');								// Print " status 0F" where 0F is the status byte
	mcu_putchar('s');
	mcu_putchar('t');
	mcu_putchar('a');
	mcu_putchar('t');
	mcu_putchar('u');
	mcu_putchar('s');
	mcu_putchar(' ');
	db_puthex(statusbyte);
	db_putenter();
}
	

// Is c an ascii representation of a hex number?
int db_is_ascii_hex(int c) {
	if ((c >= '0') & (c <= '9'))					// Numbers 0 through 9
		return 1;
	if ((c >= 'A') & (c <= 'F'))					// A through F
		return 1;
	if ((c >= 'a') & (c <= 'f') )					// a through f  
		return 1;
	return 0;
}


// Convert the characters c0, c1 into hex number h
char db_ascii_to_hex(int c0, int c1) {
	char h;
	h = 0;

	// Most significant hex number
	if ((c0 >= '0') && (c0 <= '9'))				// 0-9 -> 0-9
		h = c0 - '0';
	else if ((c0 >= 'A') && (c0 <= 'F'))			// A-F -> 10-15
		h = c0 - 'A' + 10;
	else if ((c0 >= 'a') && (c0 <= 'f'))			// a-f -> 10-15
		h = c0 - 'a' + 10;
		
	h <<= 4;										// Shift 4 MSBs up

	// Least significant hex number
	if ((c1 >= '0') && (c1 <= '9'))				 // 0-9 -> 0-9
		h = h + c1 - '0';
	else if ((c1 >= 'A') && (c1 <= 'F'))			// A-F -> 10-15
		h = h + c1 - 'A' + 10;
	else if ((c1 >= 'a') && (c1 <= 'f'))			// a-f -> 10-15
		h = h + c1 - 'a' + 10;
		
	return h;
}


// Writes a new line to the terminal
void db_putenter(void) {
	mcu_putchar('\r');
	mcu_putchar('\n');
}
	

// Convert lowest nibble to ascii character
void db_puthex(char c) {
	char temp = c;
	
	c >>=4;											// Shift in most significant hex character
	if (c < 10)										// 0-9 -> '0' - '9'
		mcu_putchar(c + '0');
	else											// A-F -> 'A' - 'F'
		mcu_putchar(c - 0x0A + 'A');
	
	c = temp & 0x0F;								// Mask in least significant hex character
	if (c < 10)										// 0-9 -> '0' - '9'
		mcu_putchar(c + '0');
	else											// A-F -> 'A' - 'F'
		mcu_putchar(c - 0x0A + 'A');
}


// Get a hex character from the terminal
char db_gethex(int c0) {							// c0 already fetched by debug interface
	extern char error;								// Global error register
	int c1;											// Input characters from the terminal
	
	error = 0;										// NO error detected yet
	if (db_is_ascii_hex(c0)) {						// If it's an ascii hex number, 
		c1 = mcu_getchar();							// Get second character
		if (db_is_ascii_hex(c1))					// If it's an ascii hex number too, 
			return db_ascii_to_hex(c0, c1);		// Give out a hex number
		else
			error = 2;								// Detected an error in 2nd character
	}
	else
		error = 1;									// Detected an error in 1st character
	
	return 0;										// A fairly stupid default value
}


// Display track, minutes and seconds FIX: add negative time indicator
void db_display_CD(char track, char min, char sec) {
	db_putenter();
	if (track == EMPTYCHAR) {						// EMPTYCHAR means that we should fill in blanks for the item
		mcu_putchar(' ');
		mcu_putchar(' ');
	}
	else
		db_puthex(track);							// Otherwise, fill in hex code. FIX: update to decimal!
	mcu_putchar(' '); 
	if (min == EMPTYCHAR) {
		mcu_putchar(' ');
		mcu_putchar(' ');
	}
	else
		db_puthex(min); 
	mcu_putchar(':'); 
	if (sec == EMPTYCHAR) {
		mcu_putchar(' ');
		mcu_putchar(' ');
	}
	else
		db_puthex(sec);
	mcu_putchar(' ');
	mcu_putchar(' ');
	mcu_putchar(' ');
}



#ifdef DEBUG_IF

// User interface to read and write raw hex data in Z1
void db_hex(void) {
	extern char statusbyte;							// Global status byte
	extern char error;								// Global error register
	
	int c0;
	char run=1;										// Keep running hex user interface!
	   
	while (run==1) {								// User interface is infinite loop
		db_putenter();
		mcu_putchar('>');
		
		c0 = mcu_getchar();							// Get command/address character
		if (c0 == 'Q') {							// "Q" for quit
			db_putenter();
			mcu_putchar('b');
			mcu_putchar('y');
			mcu_putchar('e');
			db_putenter();
			db_putenter();
			run = 0;								// Break loop
		}
		else if (c0 == 'D')						// 'D' sets D
			mcu_set_D();
		else if (c0 == 'S')						// 'S' sets S
			mcu_set_S();
		else if (c0 == 'A')						// 'A' sets A
			mcu_set_A();
		else if (c0 == 'd')						// 'd' sets D
			mcu_clear_D();
		else if (c0 == 's')						// 's' sets S
			mcu_clear_S();
		else if (c0 == 'a')						// 'a' sets A
			mcu_clear_A();
		else if (c0 == 'R') {						// 'R' prints report of DSA
			db_putenter();
			mcu_putchar('D'); mcu_putchar('S'); mcu_putchar('A'); mcu_putchar('='); 
			if (mcu_ishigh_D()) mcu_putchar('1'); else mcu_putchar('0');
			if (mcu_ishigh_S()) mcu_putchar('1'); else mcu_putchar('0');
			if (mcu_ishigh_A()) mcu_putchar('1'); else mcu_putchar('0');
			db_putenter();
		}
	} // end of infinite for
} // end of main
#endif // DEBUG_IF



#endif												// This whole file is inside a define!
